import os
import math
import torch
import torch.nn as nn
import torch.utils.model_zoo as model_zoo
from modules.img_enc import EncoderImage_ProjFeature


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1,
                               stride=stride, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    def __init__(self, block, layers, width_mult, num_classes=1000):
        super(ResNet, self).__init__()
        self.inplanes = 64 * width_mult
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7,
                               stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2,
                                    padding=0, ceil_mode=True)
        self.layer1 = self._make_layer(block, 64 * width_mult, layers[0])
        self.layer2 = self._make_layer(block, 128 * width_mult, layers[1],
                                       stride=2)
        self.layer3 = self._make_layer(block, 256 * width_mult, layers[2],
                                       stride=2)
        self.layer4 = self._make_layer(block, 512 * width_mult, layers[3],
                                       stride=2)
        self.avgpool = nn.AvgPool2d(7)
        self.fc = nn.Linear(512 * block.expansion * width_mult, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


def resnet101(pretrained=False, width_mult=1):
    model = ResNet(Bottleneck, [3, 4, 23, 3], width_mult)
    return model


class ResNetFeatureExtractor(nn.Module):
    def __init__(self, weights_path,
                 pooling_size=7, fixed_blocks=2):
        super(ResNetFeatureExtractor, self).__init__()
        self.weights_path = weights_path
        self.pooling_size = pooling_size
        self.fixed_blocks = fixed_blocks

        self.resnet = resnet101()
        self._init_modules()

    def _init_modules(self):
        # build resnet
        self.base = nn.Sequential(
            self.resnet.conv1, self.resnet.bn1, self.resnet.relu,
            self.resnet.maxpool, self.resnet.layer1, self.resnet.layer2,
            self.resnet.layer3,
        )
        self.top = nn.Sequential(self.resnet.layer4)

        if os.path.exists(self.weights_path):
            backbone_ckpt = torch.load(self.weights_path)
            self.base.load_state_dict(backbone_ckpt['base'])
            self.top.load_state_dict(backbone_ckpt['top'])

    def set_fixed_blocks(self, fixed_blocks):
        self.fixed_blocks = fixed_blocks

    def get_fixed_blocks(self):
        return self.fixed_blocks

    def unfreeze_base(self, fixed_blocks):
        print(fixed_blocks)
        assert (0 <= fixed_blocks < 4)
        if fixed_blocks == 3:
            for p in self.base[6].parameters():
                p.requires_grad = False
            for p in self.base[5].parameters():
                p.requires_grad = False
            for p in self.base[4].parameters():
                p.requires_grad = False
            for p in self.base[0].parameters():
                p.requires_grad = False
            for p in self.base[1].parameters():
                p.requires_grad = False
        if fixed_blocks == 2:
            for p in self.base[6].parameters():
                p.requires_grad = True
            for p in self.base[5].parameters():
                p.requires_grad = False
            for p in self.base[4].parameters():
                p.requires_grad = False
            for p in self.base[0].parameters():
                p.requires_grad = False
            for p in self.base[1].parameters():
                p.requires_grad = False
        if fixed_blocks == 1:
            for p in self.base[6].parameters():
                p.requires_grad = True
            for p in self.base[5].parameters():
                p.requires_grad = True
            for p in self.base[4].parameters():
                p.requires_grad = False
            for p in self.base[0].parameters():
                p.requires_grad = False
            for p in self.base[1].parameters():
                p.requires_grad = False
        if fixed_blocks == 0:
            for p in self.base[6].parameters():
                p.requires_grad = True
            for p in self.base[5].parameters():
                p.requires_grad = True
            for p in self.base[4].parameters():
                p.requires_grad = True
            for p in self.base[0].parameters():
                p.requires_grad = True
            for p in self.base[1].parameters():
                p.requires_grad = True

    def freeze_base(self):
        for p in self.base.parameters():
            p.requires_grad = False

    def train(self, mode=True):
        nn.Module.train(self, mode)
        if mode:  # train
            def set_bn_eval(m):
                classname = m.__class__.__name__
                if classname.find('BatchNorm') != -1:
                    m.eval()

            self.base.apply(set_bn_eval)
            self.top.apply(set_bn_eval)

    def forward(self, img):
        batch_size = img.size(0)
        feature = self.base(img)
        feature = self.top(feature)
        feature = feature.view(batch_size, feature.size(1),
                               -1).permute(0, 2, 1)
        return feature    